class Reader:
    def __init__(self, buffer, endian='big'):
        self.buffer = buffer
        self.endian = endian
        self.i = 0

    def read(self, length=1):
        result = self.buffer[self.i:self.i + length]
        self.i += length

        return result

    def readUInteger(self, length=1):
        result = 0
        for x in range(length):
            byte = self.buffer[self.i]

            bit_padding = x * 8
            if self.endian == 'big':
                bit_padding = (8 * (length - 1)) - bit_padding

            result |= byte << bit_padding
            self.i += 1

        return result

    def readInteger(self, length=1):
        integer = self.readUInteger(length)
        result = integer
        if integer > 2 ** (length * 8) / 2:
            result -= 2 ** (length * 8)
        return result

    def readUInt64(self):
        return self.readUInteger(8)

    def readInt64(self):
        return self.readInteger(8)

    def readFloat(self):
        as_int = self.readUInt32()
        binary = bin(as_int)
        binary = binary[2:].zfill(32)

        sign = -1 if binary[0] == '1' else 1
        exponent = int(binary[1:9], 2) - 127
        mantissa_base = binary[9:]
        mantissa_bin = '1' + mantissa_base
        mantissa = 0
        val = 1

        if exponent == -127:
            if mantissa_base[1] == -1:
                return 0
            else:
                exponent = -126
                mantissa_bin = '0' + mantissa_base

        for char in mantissa_bin:
            mantissa += val * int(char)
            val = val / 2

        result = sign * 2 ** exponent * mantissa
        return result

    def readUInt32(self):
        return self.readUInteger(4)

    def readInt32(self):
        return self.readInteger(4)

    def readNUInt16(self):
        return self.readUInt16() / 65535

    def readUInt16(self):
        return self.readUInteger(2)

    def readNInt16(self):
        return self.readInt16() / 32512

    def readInt16(self):
        return self.readInteger(2)

    def readUInt8(self):
        return self.readUInteger()

    def readInt8(self):
        return self.readInteger()

    def readBool(self):
        if self.readUInt8() >= 1:
            return True
        else:
            return False

    readUInt = readUInteger
    readInt = readInteger

    readULong = readUInt64
    readLong = readInt64

    readNUShort = readNUInt16
    readNShort = readNInt16

    readUShort = readUInt16
    readShort = readInt16

    readUByte = readUInt8
    readByte = readInt8

    def readChar(self, length=1):
        return self.read(length).decode('utf-8')

    def readString(self):
        length = self.readUShort()
        return self.readChar(length)

    def tell(self):
        return self.i



def _(*args):
    print('[ScwUtils]', end=' ')
    for arg in args:
        print(arg, end=' ')
    print()


class Writer:
    def __init__(self):
        self.writen = ''

        self.temp_vertices_offsets = {
            'POSITION': 0,
            'TEXCOORD': 0,
            'NORMAL': 0
        }

        self.vertices_offsets = {
            'POSITION': 0,
            'TEXCOORD': 0,
            'NORMAL': 0
        }

    def write(self, data: dict):
        for geom in data['geometries']:
            for key in self.vertices_offsets.keys():
                self.vertices_offsets[key] = self.temp_vertices_offsets[key]
            prefix = ''

            name = geom['name']
            vertices = geom['vertices']
            materials = geom['materials']
            for vertex in vertices:
                if vertex['type'] == 'POSITION':
                    prefix = 'v '
                elif vertex['type'] == 'NORMAL':
                    prefix = 'vn '
                elif vertex['type'] == 'TEXCOORD':
                    prefix = 'vt '

                self.temp_vertices_offsets[vertex['type']] += len(vertex['vertex'])

                for item in vertex['vertex']:
                    temp_string = prefix
                    for subitem in item:
                        temp_string += str(subitem * vertex['scale']) + ' '
                    self.writen += f'{temp_string}\n'
                self.writen += '\n\n'
            for material in materials:
                self.writen += f'o {name}_{material["name"]}\n\n'
                for item in material['polygons']:
                    temp_string = 'f '
                    for subitem in item:
                        temp_list = [
                            str(subitem[0] + self.vertices_offsets['POSITION'] + 1),  # POSITION
                            str(subitem[2] + self.vertices_offsets['TEXCOORD'] + 1),  # TEXCOORD
                            str(subitem[1] + self.vertices_offsets['NORMAL'] + 1)  # NORMAL
                        ]

                        temp_string += '/'.join(temp_list) + ' '
                    self.writen += f'{temp_string}\n'
                self.writen += '\n\n'


class Parser:
    def __init__(self, file_data: str):
        # <Variables>

        self.parsed = {'header': {'version': 2,
                                  'frame_rate': 30,
                                  'materials_file': 'sc3d/character_materials.scw'},
                       'materials': [],
                       'geometries': [],
                       'cameras': [],
                       'nodes': []}

        self.lines = file_data.split('\n')

        # <Vertices>
        self.position_temp, self.position = [], []
        self.normals_temp, self.normals = [], []
        self.texcoord_temp, self.texcoord = [], []
        # </Vertices>

        self.polygons = []

        # </Variables>

        self.parse()

    def parse(self):
        geometry_name = 'This model haven \'t a chunk_name!:( Its VERY SAD!'
        is_first_name = True
        for line in self.lines:
            items = line.split()[1:]
            if line.startswith('v '):  # POSITION
                for item in items:
                    self.position_temp.append(float(item))
            elif line.startswith('vn '):  # NORMAL
                for item in items:
                    self.normals_temp.append(float(item))
            elif line.startswith('vt '):  # TEXCOORD
                if len(items) > 2:
                    items = items[:-1]
                for item in items:
                    self.texcoord_temp.append(float(item))
            elif line.startswith('f '):
                temp_list = []
                if len(items) > 3:
                    _('It is necessary to triangulate the model')
                    break
                for item in items:
                    second_temp_list = []
                    if len(item.split('/')) == 2:
                        _('Model have not normals or texture')
                        break
                    elif len(item.split('/')) == 1:
                        _('Model have not normals and texture')
                        break
                    for x in item.split('/'):
                        second_temp_list.append(int(x) - 1)
                    temp_list.append([second_temp_list[0], second_temp_list[2], second_temp_list[1]])
                self.polygons.append(temp_list)
            elif line.startswith('o '):
                # if not is_first_name:
                #     position_scale = self.get_vertex_scale(self.position_temp)
                #     normals_scale = self.get_vertex_scale(self.normals_temp)
                #     texcoord_scale = self.get_vertex_scale(self.texcoord_temp)
                #
                #     self.parsed['geometries'].append({
                #         'chunk_name': geometry_name,
                #         'group': '',
                #         'vertices': [
                #             {'type': 'POSITION', 'index': 0, 'scale': position_scale, 'vertex': self.position},
                #             {'type': 'NORMAL', 'index': 1, 'scale': normals_scale, 'vertex': self.normals},
                #             {'type': 'TEXCOORD', 'index': 2, 'scale': texcoord_scale, 'vertex': self.texcoord}
                #         ],
                #         'have_bind_matrix': False,
                #         'materials': [{'chunk_name': 'character_mat', 'polygons': self.polygons}]
                #     })
                #
                #     # <VariablesReset>
                #
                #     # <Vertices>
                #     self.position_temp, self.position = [], []
                #     self.normals_temp, self.normals = [], []
                #     self.texcoord_temp, self.texcoord = [], []
                #     # </Vertices>
                #
                #     self.polygons = []
                #
                #     # </VariablesReset>
                # geometry_name = line.split('o ')[0]
                if is_first_name:
                    geometry_name = line.split('o ')[0]

        position_scale = self.get_vertex_scale(self.position_temp)
        normals_scale = self.get_vertex_scale(self.normals_temp)
        texcoord_scale = self.get_vertex_scale(self.texcoord_temp)

        for x in range(0, len(self.position_temp), 3):
            self.position.append(self.position_temp[x: x + 3])

        for x in range(0, len(self.normals_temp), 3):
            self.normals.append(self.normals_temp[x: x + 3])

        for x in range(0, len(self.texcoord_temp), 2):
            self.texcoord.append(self.texcoord_temp[x: x + 2])

        self.parsed['geometries'].append({
            'chunk_name': geometry_name,
            'group': '',
            'vertices': [
                {'type': 'POSITION', 'index': 0, 'scale': position_scale, 'vertex': self.position},
                {'type': 'NORMAL', 'index': 1, 'scale': normals_scale, 'vertex': self.normals},
                {'type': 'TEXCOORD', 'index': 2, 'scale': texcoord_scale, 'vertex': self.texcoord}
            ],
            'have_bind_matrix': False,
            'materials': [{'chunk_name': 'character_mat', 'polygons': self.polygons}]
        })

        # TODO: nodes

    @staticmethod
    def get_vertex_scale(vertex_data: list):
        vertex_scale = max(max(vertex_data), abs(min(vertex_data)))
        if vertex_scale < 1:
            vertex_scale = 1
        return vertex_scale



from .matrix3x3 import Matrix3x3
from . import Matrix


class Matrix4x4(Matrix):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.translation_matrix = None
        self.rotation_matrix = None
        self.scale_matrix = None

    def determinant(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]
        value1_3 = self.matrix[0][2]
        value1_4 = self.matrix[0][3]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]
        value2_3 = self.matrix[1][2]
        value2_4 = self.matrix[1][3]

        value3_1 = self.matrix[2][0]
        value3_2 = self.matrix[2][1]
        value3_3 = self.matrix[2][2]
        value3_4 = self.matrix[2][3]

        value4_1 = self.matrix[3][0]
        value4_2 = self.matrix[3][1]
        value4_3 = self.matrix[3][2]
        value4_4 = self.matrix[3][3]

        matrix1 = Matrix3x3(matrix=(
            (value2_2, value2_3, value2_4),
            (value3_2, value3_3, value3_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix2 = Matrix3x3(matrix=(
            (value2_1, value2_3, value2_4),
            (value3_1, value3_3, value3_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix3 = Matrix3x3(matrix=(
            (value2_1, value2_2, value2_4),
            (value3_1, value3_2, value3_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix4 = Matrix3x3(matrix=(
            (value2_1, value2_2, value2_3),
            (value3_1, value3_2, value3_3),
            (value4_1, value4_2, value4_3)
        ))

        det = value1_1 * matrix1.determinant()
        det -= value1_2 * matrix2.determinant()
        det += value1_3 * matrix3.determinant()
        det -= value1_4 * matrix4.determinant()

        return det

    def cofactor(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]
        value1_3 = self.matrix[0][2]
        value1_4 = self.matrix[0][3]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]
        value2_3 = self.matrix[1][2]
        value2_4 = self.matrix[1][3]

        value3_1 = self.matrix[2][0]
        value3_2 = self.matrix[2][1]
        value3_3 = self.matrix[2][2]
        value3_4 = self.matrix[2][3]

        value4_1 = self.matrix[3][0]
        value4_2 = self.matrix[3][1]
        value4_3 = self.matrix[3][2]
        value4_4 = self.matrix[3][3]

        matrix1 = Matrix3x3(matrix=(
            (value2_2, value2_3, value2_4),
            (value3_2, value3_3, value3_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix2 = Matrix3x3(matrix=(
            (value2_1, value2_3, value2_4),
            (value3_1, value3_3, value3_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix3 = Matrix3x3(matrix=(
            (value2_1, value2_2, value2_4),
            (value3_1, value3_2, value3_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix4 = Matrix3x3(matrix=(
            (value2_1, value2_2, value2_3),
            (value3_1, value3_2, value3_3),
            (value4_1, value4_2, value4_3)
        ))

        matrix5 = Matrix3x3(matrix=(
            (value1_2, value1_3, value1_4),
            (value3_2, value3_3, value3_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix6 = Matrix3x3(matrix=(
            (value1_1, value1_3, value1_4),
            (value3_1, value3_3, value3_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix7 = Matrix3x3(matrix=(
            (value1_1, value1_2, value1_4),
            (value3_1, value3_2, value3_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix8 = Matrix3x3(matrix=(
            (value1_1, value1_2, value1_3),
            (value3_1, value3_2, value3_3),
            (value4_1, value4_2, value4_3)
        ))

        matrix9 = Matrix3x3(matrix=(
            (value1_2, value1_3, value1_4),
            (value2_2, value2_3, value2_4),
            (value4_2, value4_3, value4_4)
        ))

        matrix10 = Matrix3x3(matrix=(
            (value1_1, value1_3, value1_4),
            (value2_1, value2_3, value2_4),
            (value4_1, value4_3, value4_4)
        ))

        matrix11 = Matrix3x3(matrix=(
            (value1_1, value1_2, value1_4),
            (value2_1, value2_2, value2_4),
            (value4_1, value4_2, value4_4)
        ))

        matrix12 = Matrix3x3(matrix=(
            (value1_1, value1_2, value1_3),
            (value2_1, value2_2, value2_3),
            (value4_1, value4_2, value4_3)
        ))

        matrix13 = Matrix3x3(matrix=(
            (value1_2, value1_3, value1_4),
            (value2_2, value2_3, value2_4),
            (value3_2, value3_3, value3_4)
        ))

        matrix14 = Matrix3x3(matrix=(
            (value1_1, value1_3, value1_4),
            (value2_1, value2_3, value2_4),
            (value3_1, value3_3, value3_4)
        ))

        matrix15 = Matrix3x3(matrix=(
            (value1_1, value1_2, value1_4),
            (value2_1, value2_2, value2_4),
            (value3_1, value3_2, value3_4)
        ))

        matrix16 = Matrix3x3(matrix=(
            (value1_1, value1_2, value1_3),
            (value2_1, value2_2, value2_3),
            (value3_1, value3_2, value3_3)
        ))

        self.matrix = (
            (matrix1.determinant(), matrix2.determinant(), matrix3.determinant(), matrix4.determinant()),
            (matrix5.determinant(), matrix6.determinant(), matrix7.determinant(), matrix8.determinant()),
            (matrix9.determinant(), matrix10.determinant(), matrix11.determinant(), matrix12.determinant()),
            (matrix13.determinant(), matrix14.determinant(), matrix15.determinant(), matrix16.determinant())
        )

        self.find_cofactor()

    def put_rotation(self, xyz: tuple, w: float):
        x, y, z = xyz

        rotation_matrix = (
            (1-2*y**2-2*z**2, 2*x*y-2*z*w, 2*x*z+2*y*w, 0),  # x
            (2*x*y+2*z*w, 1-2*x**2-2*z**2, 2*y*z-2*x*w, 0),  # y
            (2*x*z-2*y*w, 2*y*z+2*x*w, 1-2*x**2-2*y**2, 0),  # z
            (0, 0, 0, 1)
        )

        self.rotation_matrix = Matrix4x4(matrix=rotation_matrix)

    def put_position(self, xyz: tuple):
        x, y, z = xyz

        translation_matrix = (
            (1, 0, 0, x),  # x
            (0, 1, 0, y),  # y
            (0, 0, 1, z),  # z
            (0, 0, 0, 1)
        )

        self.translation_matrix = Matrix4x4(matrix=translation_matrix)

    def put_scale(self, xyz: tuple):
        x, y, z = xyz

        scale_matrix = (
            (x, 0, 0, 0),
            (0, y, 0, 0),
            (0, 0, z, 0),
            (0, 0, 0, 1)
        )

        self.scale_matrix = Matrix4x4(matrix=scale_matrix)

    def get_rotation(self) -> dict:
        rotation = {
            'x': 0,
            'y': 0,
            'z': 0,
            'w': 0
        }

        return rotation

    def get_position(self) -> dict:
        position = (self.matrix[0][3], self.matrix[1][3], self.matrix[2][3])

        self.put_position(position)

        position = {
            'x': position[0],
            'y': position[1],
            'z': position[2]
        }
        return position

    def get_scale(self) -> dict:
        xyz = (1, 1, 1)

        self.put_scale(xyz)

        scale = {
            'x': xyz[0],
            'y': xyz[1],
            'z': xyz[2]
        }

        return scale



class Writer:
    def __init__(self, endian='big'):
        self.endian = endian
        self.buffer = b''

    def write(self, data):
        self.buffer += data

    def writeUInteger(self, integer, length=1):
        self.buffer += integer.to_bytes(length, self.endian, signed=False)

    def writeInteger(self, integer, length=1):
        self.buffer += integer.to_bytes(length, self.endian, signed=True)

    def writeUInt64(self, integer):
        self.writeUInteger(integer, 8)

    def writeInt64(self, integer):
        self.writeInteger(integer, 8)

    def writeFloat(self, floating):
        exponent = 0
        sign = 1

        if floating == 0:
            self.writeUInt32(0)
        else:
            if floating < 0:
                sign = -1
                floating = -floating

            if floating >= 2 ** -1022:
                value = floating

                while value < 1:
                    exponent -= 1
                    value *= 2
                while value >= 2:
                    exponent += 1
                    value /= 2

            mantissa = floating / 2 ** exponent

            exponent += 127

            as_integer_bin = '0'
            if sign == -1:
                as_integer_bin = '1'

            as_integer_bin += bin(exponent)[2:].zfill(8)

            mantissa_bin = ''
            for x in range(24):
                bit = '0'
                if mantissa >= 1 / 2 ** x:
                    mantissa -= 1 / 2 ** x
                    bit = '1'
                mantissa_bin += bit

            mantissa_bin = mantissa_bin[1:]

            as_integer_bin += mantissa_bin
            as_integer = int(as_integer_bin, 2)

            self.writeUInt32(as_integer)

    def writeUInt32(self, integer):
        self.writeUInteger(integer, 4)

    def writeInt32(self, integer):
        self.writeInteger(integer, 4)

    def writeNUInt16(self, integer):
        self.writeUInt16(round(integer * 65535))

    def writeUInt16(self, integer):
        self.writeUInteger(integer, 2)

    def writeNInt16(self, integer):
        self.writeInt16(round(integer * 32512))

    def writeInt16(self, integer):
        self.writeInteger(integer, 2)

    def writeUInt8(self, integer):
        self.writeUInteger(integer)

    def writeInt8(self, integer):
        self.writeInteger(integer)

    def writeBool(self, boolean: bool):
        if boolean:
            self.writeUInt8(1)
        else:
            self.writeUInt8(0)

    writeUInt = writeUInteger
    writeInt = writeInteger

    writeULong = writeUInt64
    writeLong = writeInt64

    writeNUShort = writeNUInt16
    writeNShort = writeNInt16

    writeUShort = writeUInt16
    writeShort = writeInt16

    writeUByte = writeUInt8
    writeByte = writeInt8

    def writeChar(self, string):
        for char in list(string):
            self.buffer += char.encode('utf-8')

    def writeString(self, string):
        encoded = string.encode('utf-8')
        self.writeUShort(len(encoded))
        self.buffer += encoded



from . import Matrix


class Matrix2x2(Matrix):
    def determinant(self):
        value1_1 = self.matrix[0][0]
        value1_2 = self.matrix[0][1]

        value2_1 = self.matrix[1][0]
        value2_2 = self.matrix[1][1]

        det = value1_1 * value2_2
        det -= value1_2 * value2_1

        return det

    def cofactor(self):
        self.find_cofactor()



from ..utils.reader import Reader
from ..utils.writer import Writer
from .chunk import Chunk


class HEAD(Chunk, Writer, Reader):
    def __init__(self, header=None):
        super().__init__(header)
        self.chunk_name = 'HEAD'

    def parse(self, buffer: bytes):
        Reader.__init__(self, buffer)

        setattr(self, 'version', self.readUShort())
        setattr(self, 'frame_rate', self.readUShort())
        setattr(self, 'v1', self.readUShort())
        setattr(self, 'v2', self.readUShort())
        setattr(self, 'materials_file', self.readString())
        if self.get('version') == 2:
            setattr(self, 'v3', self.readUByte())

    def encode(self):
        Writer.__init__(self)

        self.writeUShort(2)  # getattr(self, 'version')
        self.writeUShort(getattr(self, 'frame_rate'))
        self.writeUShort(0)  # getattr(self, 'v1')
        self.writeUShort(249)  # getattr(self, 'v2')
        self.writeString(getattr(self, 'materials_file'))
        self.writeUByte(0)  # getattr(self, 'v3')

        self.length = len(self.buffer)


class MATE(Chunk, Writer, Reader):
    def __init__(self, header: dict):
        super().__init__(header)
        self.chunk_name = 'MATE'

    def parse(self, buffer: bytes):
        Reader.__init__(self, buffer)

        setattr(self, 'name', self.readString())
        setattr(self, 'shader', self.readString())
        setattr(self, 'v1', self.readUByte())

        effect = {}
        use_ambient_tex = self.readBool()
        if use_ambient_tex:
            ambient_tex = self.readString()
            effect['ambient'] = ambient_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            ambient_color = (r, g, b, a)
            effect['ambient'] = ambient_color

        use_diffuse_tex = self.readBool()
        if use_diffuse_tex:
            diffuse_tex = self.readString()
            effect['diffuse'] = diffuse_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            diffuse_color = (r, g, b, a)
            effect['diffuse'] = diffuse_color

        use_specular_tex = self.readBool()
        if use_specular_tex:
            specular_tex = self.readString()
            effect['specular'] = specular_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            specular_color = (r, g, b, a)
            effect['specular'] = specular_color

        setattr(self, 'v2', self.readString())
        setattr(self, 'v3', self.readString())

        use_colorize_tex = self.readBool()
        if use_colorize_tex:
            colorize_tex = self.readString()
            effect['colorize'] = colorize_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            colorize_color = (r, g, b, a)
            effect['colorize'] = colorize_color

        use_emission_tex = self.readBool()
        if use_emission_tex:
            emission_tex = self.readString()
            effect['emission'] = emission_tex
        else:
            a = self.readUByte()
            r = self.readUByte()
            g = self.readUByte()
            b = self.readUByte()
            emission_color = (r, g, b, a)
            effect['emission'] = emission_color

        setattr(self, 'v4', self.readString())
        setattr(self, 'v5', self.readFloat())
        setattr(self, 'v6', self.readFloat())

        effect['lightmaps'] = {
            'diffuse': self.readString(),
            'specular': self.readString()
        }

        a = self.readUByte()
        r = self.readUByte()
        g = self.readUByte()
        b = self.readUByte()
        effect['tint'] = (r, g, b, a)

        setattr(self, 'effect', effect)

    def encode(self):
        Writer.__init__(self)

        self.length = len(self.buffer)


class GEOM(Chunk, Writer, Reader):
    def __init__(self, header: dict):
        super().__init__(header)
        self.chunk_name = 'GEOM'

    def parse(self, buffer: bytes):
        Reader.__init__(self, buffer)

        setattr(self, 'name', self.readString())
        setattr(self, 'group', self.readString())
        if self.header['version'] < 2:
            matrix = []
            for x in range(4):
                temp_list = []
                for x1 in range(4):
                    temp_list.append(self.readFloat())
                matrix.append(temp_list)

        self.parse_vertices()
        self.parse_skin()
        self.parse_materials()

    def parse_vertices(self):
        vertices = []
        inputs = []

        vertex_count = self.readUByte()
        for x in range(vertex_count):
            vertex = []
            vertex_type = self.readString()
            vertex_index = self.readUByte()
            vertex_shorts = self.readUShort()
            vertex_scale = self.readFloat()
            vertex_count = self.readUInt32()

            if vertex_type == 'VERTEX':
                vertex_type = 'POSITION'

            for x1 in range(vertex_count):
                coordinates_massive = []
                for x2 in range(vertex_shorts):
                    coordinate = self.readShort()
                    coordinates_massive.append(coordinate / 32512)
                if vertex_type == 'TEXCOORD':
                    coordinates_massive[1::2] = [1 - x for x in coordinates_massive[1::2]]
                vertex.append(coordinates_massive)

            inputs.append({
                'type': vertex_type,
                'offset': x,
                'name': f'{vertex_type.lower()}_0'
            })

            vertices.append({
                'name': f'{vertex_type.lower()}_0',
                'type': vertex_type,
                'index': vertex_index,
                'scale': vertex_scale,
                'vertex': vertex
            })
        setattr(self, 'inputs', inputs)
        setattr(self, 'vertices', vertices)

    def parse_skin(self):
        bind_matrix = []

        setattr(self, 'have_bind_matrix', self.readBool())
        if getattr(self, 'have_bind_matrix'):
            for x in range(16):
                bind_matrix.append(self.readFloat())

        setattr(self, 'bind_matrix', bind_matrix)

        self.parse_joints()
        self.parse_weights()

    def parse_joints(self):
        joints = []

        joint_counts = self.readUByte()
        for x in range(joint_counts):
            joint_matrix = []
            joint_name = self.readString()
            for x1 in range(16):
                joint_matrix.append(self.readFloat())
            joints.append({'name': joint_name, 'matrix': joint_matrix})

        setattr(self, 'joints', joints)

    def parse_weights(self):
        vertex_weights = []
        weights = []
        vcount = []

        vertex_weights_count = self.readUInt32()
        for x in range(vertex_weights_count):
            vcount.append(0)
            joint_a = self.readUByte()
            joint_b = self.readUByte()
            joint_c = self.readUByte()
            joint_d = self.readUByte()
            weight_a = self.readUShort()
            weight_b = self.readUShort()
            weight_c = self.readUShort()
            weight_d = self.readUShort()
            temp_list = [
                [joint_a, weight_a],
                [joint_b, weight_b],
                [joint_c, weight_c],
                [joint_d, weight_d]
            ]
            for pair in temp_list:
                if pair[1] != 0:
                    vcount[x] += 1
                    vertex_weights.append(pair[0])
                    if pair[1] / 65535 not in weights:
                        weights.append(pair[1] / 65535)
                    vertex_weights.append(weights.index(pair[1] / 65535))

        setattr(self, 'weights',
                {
                    'vertex_weights': vertex_weights,
                    'weights': weights,
                    'vcount': vcount
                })

    def parse_materials(self):
        materials = []

        materials_count = self.readUByte()
        for x in range(materials_count):
            polygons = []
            material_name = self.readString()
            self.readString()
            polygons_count = self.readUShort()
            inputs_count = self.readUByte()
            vertex_id_length = self.readUByte()
            for x1 in range(polygons_count):
                temp_list = []
                for x2 in range(3):
                    second_temp_list = []
                    for x3 in range(inputs_count):
                        second_temp_list.append(self.readUInteger(vertex_id_length))
                    temp_list.append(second_temp_list)
                polygons.append(temp_list)
            materials.append({
                'name': material_name,
                'inputs': getattr(self, 'inputs'),
                'polygons': polygons
            })

        setattr(self, 'materials', materials)

    def encode(self):
        Writer.__init__(self)

        self.writeString(self.get('name'))
        self.writeString(self.get('group'))

        self.encode_vertices(self.get('vertices'))

        self.encode_skin()

        self.encode_materials()

        self.length = len(self.buffer)

    def encode_vertices(self, vertices: dict):
        self.writeUByte(len(vertices))
        for vertex in vertices:
            self.writeString(vertex['type'])
            self.writeUByte(vertex['index'])
            self.writeUShort(len(vertex['vertex'][0]))
            self.writeFloat(vertex['scale'])
            self.writeUInt32(len(vertex['vertex']))
            for coordinates_massive in vertex['vertex']:
                if vertex['type'] == 'TEXCOORD':
                    coordinates_massive[1::2] = [1 - x for x in coordinates_massive[1::2]]
                for coordinate in coordinates_massive:
                    # coordinate /= vertex['scale']
                    coordinate *= 32512
                    self.writeShort(round(coordinate))

    def encode_skin(self):
        self.writeBool(self.get('have_bind_matrix'))
        if self.get('have_bind_matrix'):
            for x in self.get('bind_matrix'):
                self.writeFloat(x)

        self.encode_joints()

        self.encode_weight()

    def encode_joints(self):
        if self.get('have_bind_matrix'):
            self.writeUByte(len(self.get('joints')))

            for joint in self.get('joints'):
                self.writeString(joint['name'])
                for x in joint['matrix']:
                    self.writeFloat(x)
        else:
            self.writeUByte(0)

    def encode_weight(self):
        if self.get('have_bind_matrix'):
            self.writeUInt32(len(self.get('weights')['vcount']))
            past_index = 0
            for vcount in self.get('weights')['vcount']:
                temp_list = []
                for x in range(vcount):
                    vertex_weights_index = x * 2 + past_index * 2
                    joint_id = self.get('weights')['vertex_weights'][vertex_weights_index]
                    weight_id = self.get('weights')['vertex_weights'][vertex_weights_index + 1]

                    weight = self.get('weights')['weights'][weight_id]

                    if weight > 1:
                        weight = 1
                    elif weight < 1:
                        weight = 0

                    weight = int(weight * 65535)
                    temp_list.append([joint_id, weight])
                past_index += vcount
                while len(temp_list) < 4:
                    temp_list.append([0, 0])
                for x in temp_list:
                    self.writeUByte(x[0])
                for x in temp_list:
                    self.writeUShort(x[1])
        else:
            self.writeUInt32(0)

    def encode_materials(self):
        self.writeUByte(len(self.get('materials')))
        for material in self.get('materials'):
            self.writeString(material['name'])
            self.writeString('')
            self.writeUShort(len(material['polygons']))

            # Calculate settings
            inputs_count = len(material['polygons'][0][0])

            maximal_value = 0
            for points in material['polygons']:
                for point in points:
                    for vertex in point:
                        if vertex > maximal_value:
                            maximal_value = vertex

            short_length = 1 if maximal_value <= 255 else 2

            # Write Settings
            self.writeUByte(inputs_count)
            self.writeUByte(short_length)

            # Write Polygons
            for points in material['polygons']:
                for point in points:
                    for vertex in point:
                        self.writeUInteger(vertex, short_length)


class CAME(Chunk, Writer, Reader):
    def __init__(self, header=None):
        super().__init__(header)
        self.chunk_name = 'CAME'

    def parse(self, buffer: bytes):
        super(Reader).__init__(buffer=buffer)

        setattr(self, 'name', self.readString())
        setattr(self, 'v1', self.readFloat())
        setattr(self, 'xFov', self.readFloat())
        setattr(self, 'aspectRatio', self.readFloat())
        setattr(self, 'zNear', self.readFloat())
        setattr(self, 'zFar', self.readFloat())

    def encode(self):
        Writer.__init__(self)

        self.writeString(getattr(self, 'name'))
        self.writeFloat(getattr(self, 'v1'))
        self.writeFloat(getattr(self, 'xFov'))
        self.writeFloat(getattr(self, 'aspectRatio'))
        self.writeFloat(getattr(self, 'zNear'))
        self.writeFloat(getattr(self, 'zFar'))

        self.length = len(self.buffer)


class NODE(Chunk, Writer, Reader):
    def __init__(self, header: dict):
        super().__init__(header)
        self.chunk_name = 'NODE'

    def parse(self, buffer: bytes):
        Reader.__init__(self, buffer)
        nodes = []

        nodes_count = self.readUShort()
        for node in range(nodes_count):
            node_data = {}

            node_data['name'] = self.readString()
            node_data['parent'] = self.readString()

            has_target = True if self.readUShort() else False
            node_data['has_target'] = has_target
            if has_target:
                target_type = self.readChar(4)
                target_name = self.readString()
                binds_count = self.readUShort()
                binds = []
                for bind in range(binds_count):
                    binds.append({})
                    symbol = self.readString()
                    target = self.readString()
                    binds[bind] = {'symbol': symbol,
                                   'target': target}
                node_data['target_type'] = target_type
                node_data['target'] = target_name
                node_data['binds'] = binds

            frames_count = self.readUShort()
            node_data['frames'] = []
            if frames_count > 0:
                settings = list(bin(self.readUByte())[2:].zfill(8))
                settings = [bool(int(value)) for value in settings]
                node_data['frames_settings'] = settings
                for frame in range(frames_count):
                    frame_data = {}

                    frame_data['frame_id'] = self.readUShort()
                    if settings[7] or frame == 0:  # Rotation
                        rotation = {
                            'x': self.readNShort(),
                            'y': self.readNShort(),
                            'z': self.readNShort(),
                            'w': self.readNShort()
                        }

                    if settings[4] or frame == 0:  # Position X
                        pos_x = self.readFloat()
                    if settings[5] or frame == 0:  # Position Y
                        pos_y = self.readFloat()
                    if settings[6] or frame == 0:  # Position Z
                        pos_z = self.readFloat()

                    if settings[1] or frame == 0:  # Scale X
                        scale_x = self.readFloat()
                    if settings[2] or frame == 0:  # Scale Y
                        scale_y = self.readFloat()
                    if settings[3] or frame == 0:  # Scale Z
                        scale_z = self.readFloat()

                    frame_data['rotation'] = rotation
                    frame_data['position'] = {
                        'x': pos_x,
                        'y': pos_y,
                        'z': pos_z
                    }
                    frame_data['scale'] = {
                        'x': scale_x,
                        'y': scale_y,
                        'z': scale_z
                    }

                    node_data['frames'].append(frame_data)
            nodes.append(node_data)
        setattr(self, 'nodes', nodes)

    def encode(self):
        Writer.__init__(self)

        self.writeUShort(len(self.get('nodes')))
        for node in self.get('nodes'):
            self.writeString(node['name'])
            self.writeString(node['parent'])

            self.writeUShort(1 if node['has_target'] else 0)
            if node['has_target']:
                self.writeChar(node['target_type'])
                self.writeString(node['target'])
                self.writeUShort(len(node['binds']))
                for bind in node['binds']:
                    self.writeString(bind['symbol'])
                    self.writeString(bind['target'])

            if 'frames_settings' in node:
                frames_settings = node['frames_settings']
            else:
                frames_settings = None
            self.encode_frames(node['frames'], frames_settings)

        self.length = len(self.buffer)

    def encode_frames(self, frames, frames_settings):
        self.writeUShort(len(frames))
        if len(frames) > 0:
            self.writeUByte(int(''.join([('1' if item else '0') for item in frames_settings])[::], 2))
            for frame in frames:
                self.writeUShort(frame['frame_id'])
                if frames_settings[7] or frames.index(frame) == 0:  # Rotation
                    self.writeNShort(frame['rotation']['x'])
                    self.writeNShort(frame['rotation']['y'])
                    self.writeNShort(frame['rotation']['z'])
                    self.writeNShort(frame['rotation']['w'])

                if frames_settings[4] or frames.index(frame) == 0:  # Position X
                    self.writeFloat(frame['position']['x'])
                if frames_settings[5] or frames.index(frame) == 0:  # Position Y
                    self.writeFloat(frame['position']['y'])
                if frames_settings[6] or frames.index(frame) == 0:  # Position Z
                    self.writeFloat(frame['position']['z'])

                if frames_settings[1] or frames.index(frame) == 0:  # Scale X
                    self.writeFloat(frame['scale']['x'])
                if frames_settings[2] or frames.index(frame) == 0:  # Scale Y
                    self.writeFloat(frame['scale']['y'])
                if frames_settings[3] or frames.index(frame) == 0:  # Scale Z
                    self.writeFloat(frame['scale']['z'])


class WEND(Chunk, Writer, Reader):
    def __init__(self, header=None):
        super().__init__(header)
        self.chunk_name = 'WEND'

    def parse(self, buffer: bytes):
        Reader.__init__(self, buffer)

    def encode(self):
        Writer.__init__(self)

        self.length = len(self.buffer)


__all__ = [
    'HEAD',
    'MATE',
    'GEOM',
    'CAME',
    'NODE',
    'WEND'
]
